"""Collection of plots to evaluate a model, create overviews on data or forecasts."""
__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2020-11-23'

import logging
import math
import os
import warnings
from typing import Dict, List, Tuple

import matplotlib
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
import seaborn as sns
import xarray as xr
from matplotlib.backends.backend_pdf import PdfPages

from mlair import helpers
from mlair.data_handler.iterator import DataCollection
from mlair.helpers import TimeTrackingWrapper

logging.getLogger('matplotlib').setLevel(logging.WARNING)


# import matplotlib
# matplotlib.use("TkAgg")
# import matplotlib.pyplot as plt


class AbstractPlotClass:
    """
    Abstract class for all plotting routines to unify plot workflow.

    Each inheritance requires a _plot method. Create a plot class like:

    .. code-block:: python

        class MyCustomPlot(AbstractPlotClass):

            def __init__(self, plot_folder, *args, **kwargs):
                super().__init__(plot_folder, "custom_plot_name")
                self._data = self._prepare_data(*args, **kwargs)
                self._plot(*args, **kwargs)
                self._save()

            def _prepare_data(*args, **kwargs):
                <your custom data preparation>
                return data

            def _plot(*args, **kwargs):
                <your custom plotting without saving>

    The save method is already implemented in the AbstractPlotClass. If special saving is required (e.g. if you are
    using pdfpages), you need to overwrite it. Plots are saved as .pdf with a resolution of 500dpi per default (can be
    set in super class initialisation).

    Methods like the shown _prepare_data() are optional. The only method required to implement is _plot.

    If you want to add a time tracking module, just add the TimeTrackingWrapper as decorator around your custom plot
    class. It will log the spent time if you call your plotting without saving the returned object.

    .. code-block:: python

        @TimeTrackingWrapper
        class MyCustomPlot(AbstractPlotClass):
            pass

    Let's assume it takes a while to create this very special plot.

    >>> MyCustomPlot()
    INFO: MyCustomPlot finished after 00:00:11 (hh:mm:ss)

    """

    def __init__(self, plot_folder, plot_name, resolution=500, rc_params=None):
        """Set up plot folder and name, and plot resolution (default 500dpi)."""
        plot_folder = os.path.abspath(plot_folder)
        if not os.path.exists(plot_folder):
            os.makedirs(plot_folder)
        self.plot_folder = plot_folder
        self.plot_name = plot_name
        self.resolution = resolution
        if rc_params is None:
            rc_params = {'axes.labelsize': 'large',
                         'xtick.labelsize': 'large',
                         'ytick.labelsize': 'large',
                         'legend.fontsize': 'large',
                         'axes.titlesize': 'large',
                         }
        self.rc_params = rc_params
        self._update_rc_params()

    def _plot(self, *args):
        """Abstract plot class needs to be implemented in inheritance."""
        raise NotImplementedError

    def _save(self, **kwargs):
        """Store plot locally. Name of and path to plot need to be set on initialisation."""
        plot_name = os.path.join(self.plot_folder, f"{self.plot_name}.pdf")
        logging.debug(f"... save plot to {plot_name}")
        plt.savefig(plot_name, dpi=self.resolution, **kwargs)
        plt.close('all')

    def _update_rc_params(self):
        plt.rcParams.update(self.rc_params)

    @staticmethod
    def _get_sampling(sampling):
        if sampling == "daily":
            return "D"
        elif sampling == "hourly":
            return "h"

    @staticmethod
    def get_dataset_colors():
        """
        Standard colors used for train-, val-, and test-sets during postprocessing
        """
        colors = {"train": "#e69f00", "val": "#009e73", "test": "#56b4e9"}  # hex code
        return colors


@TimeTrackingWrapper
class PlotMonthlySummary(AbstractPlotClass):
    """
    Show a monthly summary over all stations for each lead time ("ahead") as box and whiskers plot.

    The plot is saved in data_path with name monthly_summary_box_plot.pdf and 500dpi resolution.

    .. image:: ../../../../../_source/_plots/monthly_summary_box_plot.png
        :width: 400

    :param stations: all stations to plot
    :param data_path: path, where the data is located
    :param name: full name of the local files with a % as placeholder for the station name
    :param target_var: display name of the target variable on plot's axis
    :param window_lead_time: lead time to plot, if window_lead_time is higher than the available lead time or not given
        the maximum lead time from data is used. (default None -> use maximum lead time from data).
    :param plot_folder: path to save the plot (default: current directory)
    :param target_var_unit: unit of target var for plot legend (default= ppb)

    """

    def __init__(self, stations: List, data_path: str, name: str, target_var: str, window_lead_time: int = None,
                 plot_folder: str = ".", target_var_unit: str = 'ppb'):
        """Set attributes and create plot."""
        super().__init__(plot_folder, "monthly_summary_box_plot")
        self._data_path = data_path
        self._data_name = name
        self._data = self._prepare_data(stations)
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._plot(target_var, target_var_unit)
        self._save()

    def _prepare_data(self, stations: List) -> xr.DataArray:
        """
        Pre.process data required to plot.

        For each station, load locally saved predictions, extract the CNN prediction and the observation and group them
        into monthly bins (no aggregation, only sorting them).

        :param stations: all stations to plot
        :return: The entire data set, flagged with the corresponding month.
        """
        forecasts = None
        for station in stations:
            logging.debug(f"... preprocess station {station}")
            file_name = os.path.join(self._data_path, self._data_name % station)
            data = xr.open_dataarray(file_name)

            data_cnn = data.sel(type="CNN").squeeze()
            if len(data_cnn.shape) > 1:
                data_cnn = data_cnn.assign_coords(ahead=[f"{days}d" for days in data_cnn.coords["ahead"].values])

            data_obs = data.sel(type="obs", ahead=1).squeeze()
            data_obs.coords["ahead"] = "obs"

            data_concat = xr.concat([data_obs, data_cnn], dim="ahead")
            data_concat = data_concat.drop_vars("type")

            new_index = data_concat.index.values.astype("datetime64[M]").astype(int) % 12 + 1
            data_concat = data_concat.assign_coords(index=new_index)
            data_concat = data_concat.clip(min=0)

            forecasts = xr.concat([forecasts, data_concat], 'index') if forecasts is not None else data_concat
        return forecasts

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments.

        If window_lead_time is not given, extract this information from data itself by the number of ahead dimensions.
        If given, check if data supports the give length. If the number of ahead dimensions in data is lower than the
        given lead time, data's lead time is used.

        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._data.ahead)
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    @staticmethod
    def _spell_out_chemical_concentrations(short_name: str):
        short2long = {'o3': 'ozone', 'no': 'nitrogen oxide', 'no2': 'nitrogen dioxide', 'nox': 'nitrogen dioxides'}
        return f"{short2long[short_name]} concentration"

    def _plot(self, target_var: str, target_var_unit: str):
        """
        Create a monthly grouped box plot over all stations but with separate boxes for each lead time step.

        :param target_var: display name of the target variable on plot's axis
        """
        data = self._data.to_dataset(name='values').to_dask_dataframe()
        logging.debug("... start plotting")
        color_palette = [matplotlib.colors.cnames["green"]] + sns.color_palette("Blues_d",
                                                                                self._window_lead_time).as_hex()
        ax = sns.boxplot(x='index', y='values', hue='ahead', data=data.compute(), whis=1., palette=color_palette,
                         flierprops={'marker': '.', 'markersize': 1}, showmeans=True,
                         meanprops={'markersize': 1, 'markeredgecolor': 'k'})
        ylabel = self._spell_out_chemical_concentrations(target_var)
        ax.set(xlabel='month', ylabel=f'{ylabel} (in {target_var_unit})')
        plt.tight_layout()


@TimeTrackingWrapper
class PlotStationMap(AbstractPlotClass):
    """
    Plot geographical overview of all used stations as squares.

    Different data sets can be colorised by its key in the input dictionary generators. The key represents the color to
    plot on the map. Currently, there is only a white background, but this can be adjusted by loading locally stored
    topography data (not implemented yet). The plot is saved under plot_path with the name station_map.pdf

    .. image:: ../../../../../_source/_plots/station_map.png
        :width: 400
    """

    def __init__(self, generators: Dict, plot_folder: str = "."):
        """
        Set attributes and create plot.

        :param generators: dictionary with the plot color of each data set as key and the generator containing all stations
        as value.
        :param plot_folder: path to save the plot (default: current directory)
        """
        super().__init__(plot_folder, "station_map")
        self._ax = None
        self._gl = None
        self._plot(generators)
        self._save()

    def _draw_background(self):
        """Draw coastline, lakes, ocean, rivers and country borders as background on the map."""

        import cartopy.feature as cfeature

        self._ax.add_feature(cfeature.LAND.with_scale("50m"))
        self._ax.natural_earth_shp(resolution='50m')
        self._ax.add_feature(cfeature.COASTLINE.with_scale("50m"), edgecolor='black')
        self._ax.add_feature(cfeature.LAKES.with_scale("50m"))
        self._ax.add_feature(cfeature.OCEAN.with_scale("50m"))
        self._ax.add_feature(cfeature.RIVERS.with_scale("50m"))
        self._ax.add_feature(cfeature.BORDERS.with_scale("50m"), facecolor='none', edgecolor='black')

    def _plot_stations(self, generators):
        """
        Loop over all keys in generators dict and its containing stations and plot the stations's position.

        Position is highlighted by a square on the map regarding the given color.

        :param generators: dictionary with the plot color of each data set as key and the generator containing all
            stations as value.
        """

        import cartopy.crs as ccrs
        if generators is not None:
            for color, data_collection in generators.items():
                for station in data_collection:
                    coords = station.get_coordinates()
                    IDx, IDy = coords["lon"], coords["lat"]
                    self._ax.plot(IDx, IDy, mfc=color, mec='k', marker='s', markersize=6, transform=ccrs.PlateCarree())

    def _plot(self, generators: Dict):
        """
        Create the station map plot.

        Set figure and call all required sub-methods.

        :param generators: dictionary with the plot color of each data set as key and the generator containing all
            stations as value.
        """

        import cartopy.crs as ccrs
        from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
        fig = plt.figure(figsize=(10, 5))
        self._ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        self._gl = self._ax.gridlines(xlocs=range(0, 21, 5), ylocs=range(44, 59, 2), draw_labels=True)
        self._gl.xformatter = LONGITUDE_FORMATTER
        self._gl.yformatter = LATITUDE_FORMATTER
        self._draw_background()
        self._plot_stations(generators)
        self._adjust_extent()
        plt.tight_layout()

    def _adjust_extent(self):
        import cartopy.crs as ccrs

        def diff(arr):
            return arr[1] - arr[0], arr[3] - arr[2]

        def find_ratio(delta, reference=5):
            return max(abs(reference / delta[0]), abs(reference / delta[1]))

        extent = self._ax.get_extent(crs=ccrs.PlateCarree())
        ratio = find_ratio(diff(extent))
        new_extent = extent + np.array([-1, 1, -1, 1]) * ratio
        self._ax.set_extent(new_extent, crs=ccrs.PlateCarree())


@TimeTrackingWrapper
class PlotConditionalQuantiles(AbstractPlotClass):
    """
    Create cond.quantile plots as originally proposed by Murphy, Brown and Chen (1989) [But in log scale].

    Link to paper: https://journals.ametsoc.org/doi/pdf/10.1175/1520-0434%281989%29004%3C0485%3ADVOTF%3E2.0.CO%3B2

    .. image:: ../../../../../_source/_plots/conditional_quantiles_cali-ref_plot.png
        :width: 400

    .. image:: ../../../../../_source/_plots/conditional_quantiles_like-bas_plot.png
        :width: 400

    For each time step ahead a separate plot is created. If parameter plot_per_season is true, data is split by season
    and conditional quantiles are plotted for each season in addition.

    :param stations: all stations to plot
    :param data_pred_path: path to dir which contains the forecasts as .nc files
    :param plot_folder: path where the plots are stored
    :param plot_per_seasons: if `True' create cond. quantile plots for _seasons (DJF, MAM, JJA, SON) individually
    :param rolling_window: smoothing of quantiles (3 is used by Murphy et al.)
    :param model_mame: name of the model prediction as stored in netCDF file (for example "CNN")
    :param obs_name: name of observation as stored in netCDF file (for example "obs")
    :param kwargs: Some further arguments which are listed in self._opts
    """

    # ignore warnings if nans appear in quantile grouping
    warnings.filterwarnings("ignore", message="All-NaN slice encountered")
    # ignore warnings if mean is calculated on nans
    warnings.filterwarnings("ignore", message="Mean of empty slice")
    # ignore warnings for y tick = 0 on log scale (instead of 0.00001 or similar)
    warnings.filterwarnings("ignore", message="Attempted to set non-positive bottom ylim on a log-scaled axis.")

    def __init__(self, stations: List, data_pred_path: str, plot_folder: str = ".", plot_per_seasons=True,
                 rolling_window: int = 3, model_mame: str = "CNN", obs_name: str = "obs", **kwargs):
        """Initialise."""
        super().__init__(plot_folder, "conditional_quantiles")
        self._data_pred_path = data_pred_path
        self._stations = stations
        self._rolling_window = rolling_window
        self._model_name = model_mame
        self._obs_name = obs_name
        self._opts = self._get_opts(kwargs)
        self._seasons = ['DJF', 'MAM', 'JJA', 'SON'] if plot_per_seasons is True else ""
        self._data = self._load_data()
        self._bins = self._get_bins_from_rage_of_data()
        self._plot()

    @staticmethod
    def _get_opts(kwargs):
        """Extract options from kwargs."""
        return {"q": kwargs.get("q", [.1, .25, .5, .75, .9]),
                "linetype": kwargs.get("linetype", [':', '-.', '--', '-.', ':']),
                "legend": kwargs.get("legend", ['.10th and .90th quantile', '.25th and .75th quantile',
                                                '.50th quantile', 'reference 1:1']),
                "data_unit": kwargs.get("data_unit", "ppb"), }

    def _load_data(self) -> xr.DataArray:
        """
        Load plot data.

        :return: plot data
        """
        logging.debug("... load data")
        data_collector = []
        for station in self._stations:
            file = os.path.join(self._data_pred_path, f"forecasts_{station}_test.nc")
            data_tmp = xr.open_dataarray(file)
            data_collector.append(data_tmp.loc[:, :, [self._model_name, self._obs_name]].assign_coords(station=station))
        res = xr.concat(data_collector, dim='station').transpose('index', 'type', 'ahead', 'station')
        return res

    def _segment_data(self, data: xr.DataArray, x_model: str) -> xr.DataArray:
        """
        Segment data into bins.

        :param data: data to segment
        :param x_model: name of x dimension

        :return: segmented data
        """
        logging.debug("... segment data")
        # combine index and station to multi index
        data = data.stack(z=['index', 'station'])
        # replace multi index by simple position index (order is not relevant anymore)
        data.coords['z'] = range(len(data.coords['z']))
        # segment data of x_model into bins
        data.loc[x_model, ...] = data.loc[x_model, ...].to_pandas().T.apply(pd.cut, bins=self._bins,
                                                                            labels=self._bins[1:]).T.values
        return data

    @staticmethod
    def _labels(plot_type: str, data_unit: str = "ppb") -> Tuple[str, str]:
        """
        Assign (x,y) labels to plots correctly, depending on like-base or cali-ref factorization.

        :param plot_type: type of plot, either `obs` or a model name
        :param data_unit: unit of data to add to labels (default ppb)

        :return: tuple with y and x labels
        """
        names = (f"forecast concentration (in {data_unit})", f"observed concentration (in {data_unit})")
        if plot_type == "obs":
            return names
        else:
            return names[::-1]

    def _get_bins_from_rage_of_data(self) -> np.ndarray:
        """
        Get array of bins to use for quantiles.

        :return: range from 0 to data's maximum + 1 (rounded down)
        """
        return np.arange(0, math.ceil(self._data.max().max()) + 1, 1).astype(int)

    def _create_quantile_panel(self, data: xr.DataArray, x_model: str, y_model: str) -> xr.DataArray:
        """
        Calculate quantiles.

        :param data: data to calculate quantiles
        :param x_model: name of x dimension
        :param y_model: name of y dimension

        :return: quantile panel with binned data
        """
        logging.debug("... create quantile panel")
        # create empty xarray with dims: time steps ahead, quantiles, bin index (numbers create in previous step)
        quantile_panel = xr.DataArray(
            np.full([data.ahead.shape[0], len(self._opts["q"]), self._bins[1:].shape[0]], np.nan),
            coords=[data.ahead, self._opts["q"], self._bins[1:]], dims=['ahead', 'quantiles', 'categories'])
        # ensure that the coordinates are in the right order
        quantile_panel = quantile_panel.transpose('ahead', 'quantiles', 'categories')
        # calculate for each bin of the pred_name data the quantiles of the ref_name data
        for bin in self._bins[1:]:
            mask = (data.loc[x_model, ...] == bin)
            quantile_panel.loc[..., bin] = data.loc[y_model, ...].where(mask).quantile(self._opts["q"], dim=['z']).T
        return quantile_panel

    @staticmethod
    def add_affix(affix: str) -> str:
        """
        Add additional information to plot name with leading underscore or add empty string if affix is empty.

        :param affix: string to add

        :return: affix with leading underscore or empty string.
        """
        return f"_{affix}" if len(affix) > 0 else ""

    def _prepare_plots(self, data: xr.DataArray, x_model: str, y_model: str) -> Tuple[xr.DataArray, xr.DataArray]:
        """
        Get segmented data and quantile panel.

        :param data: plot data
        :param x_model: name of x dimension
        :param y_model: name of y dimension

        :return: segmented data and quantile panel
        """
        segmented_data = self._segment_data(data, x_model)
        quantile_panel = self._create_quantile_panel(segmented_data, x_model, y_model)
        return segmented_data, quantile_panel

    def _plot(self):
        """Start plotting routines: overall plot and seasonal (if enabled)."""
        logging.info(
            f"start plotting {self.__class__.__name__}, scheduled number of plots: {(len(self._seasons) + 1) * 2}")

        if len(self._seasons) > 0:
            self._plot_seasons()
        self._plot_all()

    def _plot_seasons(self):
        """Create seasonal plots."""
        for season in self._seasons:
            self._plot_base(data=self._data.where(self._data['index.season'] == season), x_model=self._model_name,
                            y_model=self._obs_name, plot_name_affix="cali-ref", season=season)
            self._plot_base(data=self._data.where(self._data['index.season'] == season), x_model=self._obs_name,
                            y_model=self._model_name, plot_name_affix="like-base", season=season)

    def _plot_all(self):
        """Plot overall conditional quantiles on full data."""
        self._plot_base(data=self._data, x_model=self._model_name, y_model=self._obs_name, plot_name_affix="cali-ref")
        self._plot_base(data=self._data, x_model=self._obs_name, y_model=self._model_name, plot_name_affix="like-base")

    @TimeTrackingWrapper
    def _plot_base(self, data: xr.DataArray, x_model: str, y_model: str, plot_name_affix: str, season: str = ""):
        """
        Create conditional quantile plots.

        :param data: data which is used to create cond. quantile plot
        :param x_model: name of model on x axis (can also be obs)
        :param y_model: name of model on y axis (can also be obs)
        :param plot_name_affix: should be `cali-ref' or `like-base'
        :param season: List of _seasons to use
        """
        segmented_data, quantile_panel = self._prepare_plots(data, x_model, y_model)
        ylabel, xlabel = self._labels(x_model, self._opts["data_unit"])
        plot_name = f"{self.plot_name}{self.add_affix(season)}{self.add_affix(plot_name_affix)}_plot.pdf"
        plot_path = os.path.join(os.path.abspath(self.plot_folder), plot_name)
        pdf_pages = matplotlib.backends.backend_pdf.PdfPages(plot_path)
        logging.debug(f"... plot path is {plot_path}")

        # create plot for each time step ahead
        y2_max = 0
        for iteration, d in enumerate(segmented_data.ahead):
            logging.debug(f"... plotting {d.values} time step(s) ahead")
            # plot smoothed lines with rolling mean
            smooth_data = quantile_panel.loc[d, ...].rolling(categories=self._rolling_window,
                                                             center=True).mean().to_pandas().T
            ax = smooth_data.plot(style=self._opts["linetype"], color='black', legend=False)
            ax2 = ax.twinx()
            # add reference line
            ax.plot([0, self._bins.max()], [0, self._bins.max()], color='k', label='reference 1:1', linewidth=.8)
            # add histogram of the segmented data (pred_name)
            handles, labels = ax.get_legend_handles_labels()
            segmented_data.loc[x_model, d, :].to_pandas().hist(bins=self._bins, ax=ax2, color='k', alpha=.3, grid=False,
                                                               rwidth=1)
            # add legend
            plt.legend(handles[:3] + [handles[-1]], self._opts["legend"], loc='upper left', fontsize='large')
            # adjust limits and set labels
            ax.set(xlim=(0, self._bins.max()), ylim=(0, self._bins.max()))
            ax.set_xlabel(xlabel, fontsize='x-large')
            ax.tick_params(axis='x', which='major', labelsize=15)
            ax.set_ylabel(ylabel, fontsize='x-large')
            ax.tick_params(axis='y', which='major', labelsize=15)
            ax2.yaxis.label.set_color('gray')
            ax2.tick_params(axis='y', colors='gray')
            ax2.yaxis.labelpad = -15
            ax2.set_yscale('log')
            if iteration == 0:
                y2_max = ax2.get_ylim()[1] + 100
            ax2.set(ylim=(0, y2_max * 10 ** 8), yticks=np.logspace(0, 4, 5))
            ax2.set_ylabel('              sample size', fontsize='x-large')
            ax2.tick_params(axis='y', which='major', labelsize=15)
            # set title and save current figure
            title = f"{d.values} time step(s) ahead{f' ({season})' if len(season) > 0 else ''}"
            plt.title(title)
            pdf_pages.savefig()
        # close all open figures / plots
        pdf_pages.close()
        plt.close('all')


@TimeTrackingWrapper
class PlotClimatologicalSkillScore(AbstractPlotClass):
    """
    Create plot of climatological skill score after Murphy (1988) as box plot over all stations.

    A forecast time step (called "ahead") is separately shown to highlight the differences for each prediction time
    step. Either each single term is plotted (score_only=False) or only the resulting scores CASE I to IV are displayed
    (score_only=True, default). Y-axis is adjusted following the data and not hard coded. The plot is saved under
    plot_folder path with name skill_score_clim_{extra_name_tag}{model_setup}.pdf and resolution of 500dpi.

    .. image:: ../../../../../_source/_plots/skill_score_clim_all_terms_CNN.png
        :width: 400

    .. image:: ../../../../../_source/_plots/skill_score_clim_CNN.png
        :width: 400

    :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
    :param plot_folder: path to save the plot (default: current directory)
    :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms (default True)
    :param extra_name_tag: additional tag that can be included in the plot name (default "")
    :param model_setup: architecture type to specify plot name (default "CNN")

    """

    def __init__(self, data: Dict, plot_folder: str = ".", score_only: bool = True, extra_name_tag: str = "",
                 model_setup: str = ""):
        """Initialise."""
        super().__init__(plot_folder, f"skill_score_clim_{extra_name_tag}{model_setup}")
        self._labels = None
        self._data = self._prepare_data(data, score_only)
        self._plot(score_only)
        self._save()

    def _prepare_data(self, data: Dict, score_only: bool) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant.

        In any case, transform data to a plot friendly format. Also set plot labels depending on the lead time
        dimensions.

        :param data: dictionary with station names as keys and 2D xarrays as values
        :param score_only: if true only scores of CASE I to IV are relevant
        :return: pre-processed data set
        """
        data = helpers.dict_to_xarray(data, "station")
        self._labels = [str(i) + "d" for i in data.coords["ahead"].values]
        if score_only:
            data = data.loc[:, ["CASE I", "CASE II", "CASE III", "CASE IV"], :]
        return data.to_dataframe("data").reset_index(level=[0, 1, 2])

    def _label_add(self, score_only: bool):
        """
        Add the phrase "terms and " if score_only is disabled or empty string (if score_only=True).

        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self, score_only):
        """
        Plot climatological skill score.

        :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms
        """
        fig, ax = plt.subplots()
        if not score_only:
            fig.set_size_inches(11.7, 8.27)
        sns.boxplot(x="terms", y="data", hue="ahead", data=self._data, ax=ax, whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, flierprops={"marker": "."})
        ax.axhline(y=0, color="grey", linewidth=.5)
        ax.set(ylabel=f"{self._label_add(score_only)}skill score", xlabel="", title="summary of all stations")
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()


@TimeTrackingWrapper
class PlotCompetitiveSkillScore(AbstractPlotClass):
    """
    Create competitive skill score plot.

    Create this plot for the given model setup and the reference models ordinary least squared ("ols") and the
    persistence forecast ("persi") for all lead times ("ahead"). The plot is saved under plot_folder with the name
    skill_score_competitive_{model_setup}.pdf and resolution of 500dpi.

    .. image:: ../../../../../_source/_plots/skill_score_competitive.png
        :width: 400

    :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
        calculated comparisons for cnn, persistence and ols.
    :param plot_folder: path to save the plot (default: current directory)
    :param model_setup: architecture type (default "CNN")

    """

    def __init__(self, data: pd.DataFrame, plot_folder=".", model_setup="CNN"):
        """Initialise."""
        super().__init__(plot_folder, f"skill_score_competitive_{model_setup}")
        self._labels = None
        self._data = self._prepare_data(data)
        self._plot()
        self._save()

    def _prepare_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Reformat given data and create plot labels and introduce the dimensions stations and comparison.

        :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
            calculated comparisons for cnn, persistence and ols.
        :return: processed data
        """
        data = pd.concat(data, axis=0)
        data = xr.DataArray(data, dims=["stations", "ahead"]).unstack("stations")
        data = data.rename({"stations_level_0": "stations", "stations_level_1": "comparison"})
        data = data.to_dataframe("data").unstack(level=1).swaplevel()
        data.columns = data.columns.levels[1]
        self._labels = [str(i) + "d" for i in data.index.levels[1].values]
        return data.stack(level=0).reset_index(level=2, drop=True).reset_index(name="data")

    def _plot(self):
        """Plot skill scores of the comparisons cnn-persi, ols-persi and cnn-ols."""
        fig, ax = plt.subplots()
        sns.boxplot(x="comparison", y="data", hue="ahead", data=self._data, whis=1., ax=ax, palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 3, "markeredgecolor": "k"}, flierprops={"marker": "."},
                    order=["cnn-persi", "ols-persi", "cnn-ols"])
        ax.axhline(y=0, color="grey", linewidth=.5)

        ax.set(ylabel="skill score", xlabel="competing models", title="summary of all stations", ylim=self._ylim())
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()

    def _ylim(self) -> Tuple[float, float]:
        """
        Calculate y-axis limits from data.

        Lower limit is the minimum of 0 and data's minimum (reduced by small subtrahend) and upper limit is data's
        maximum (increased by a small addend).

        :return:
        """
        lower = np.min([0, helpers.float_round(self._data.min()[2], 2) - 0.1])
        upper = helpers.float_round(self._data.max()[2], 2) + 0.1
        return lower, upper


@TimeTrackingWrapper
class PlotBootstrapSkillScore(AbstractPlotClass):
    """
    Create plot of climatological skill score after Murphy (1988) as box plot over all stations.

    A forecast time step (called "ahead") is separately shown to highlight the differences for each prediction time
    step. Either each single term is plotted (score_only=False) or only the resulting scores CASE I to IV are displayed
    (score_only=True, default). Y-axis is adjusted following the data and not hard coded. The plot is saved under
    plot_folder path with name skill_score_clim_{extra_name_tag}{model_setup}.pdf and resolution of 500dpi.

    By passing a list `separate_vars` containing variable names, a second plot is created showing the `separate_vars`
    and the remaining variables side by side with different scaling.

    .. image:: ../../../../../_source/_plots/skill_score_bootstrap.png
        :width: 400

    .. image:: ../../../../../_source/_plots/skill_score_bootstrap_separated.png
        :width: 400

    """

    def __init__(self, data: Dict, plot_folder: str = ".", model_setup: str = "", separate_vars: List = None):
        """
        Set attributes and create plot.

        :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
        :param plot_folder: path to save the plot (default: current directory)
        :param model_setup: architecture type to specify plot name (default "CNN")
        :param separate_vars: variables to plot separated (default: ['o3'])
        """
        super().__init__(plot_folder, f"skill_score_bootstrap_{model_setup}")
        if separate_vars is None:
            separate_vars = ['o3']
        self._labels = None
        self._x_name = "boot_var"
        self._data = self._prepare_data(data)
        self._plot()
        self._save()
        self.plot_name += '_separated'
        self._plot(separate_vars=separate_vars)
        self._save(bbox_inches='tight')

    def _prepare_data(self, data: Dict) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant.

        In any case, transform data to a plot friendly format. Also set plot labels depending on the lead time
        dimensions.

        :param data: dictionary with station names as keys and 2D xarrays as values
        :return: pre-processed data set
        """
        data = helpers.dict_to_xarray(data, "station").sortby(self._x_name)
        new_boot_coords = self._return_vars_without_number_tag(data.coords['boot_var'].values, split_by='_', keep=1)
        data = data.assign_coords({'boot_var': new_boot_coords})
        self._labels = [str(i) + "d" for i in data.coords["ahead"].values]
        if "station" not in data.dims:
            data = data.expand_dims("station")
        return data.to_dataframe("data").reset_index(level=[0, 1, 2])

    def _return_vars_without_number_tag(self, values, split_by, keep):
        arr = np.array([v.split(split_by) for v in values])
        num = arr[:, 0]
        new_val = arr[:, keep]
        if self._all_values_are_equal(num, axis=0):
            return new_val
        else:
            raise NotImplementedError


    @staticmethod
    def _all_values_are_equal(arr, axis=0):
        if np.all(arr == arr[0], axis=axis):
            return True
        else:
            return False

    def _label_add(self, score_only: bool):
        """
        Add the phrase "terms and " if score_only is disabled or empty string (if score_only=True).

        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self, separate_vars=None):
        """Plot climatological skill score."""
        if separate_vars is None:
            self._plot_all_variables()
        else:
            self._plot_selected_variables(separate_vars)

    def _plot_selected_variables(self, separate_vars: List):
        # if separate_vars is None:
        #     separate_vars = ['o3']
        data = self._data
        self.raise_error_if_separate_vars_do_not_exist(data, separate_vars)
        all_variables = self._get_unique_values_from_column_of_df(data, 'boot_var')
        # remaining_vars = helpers.list_pop(all_variables, separate_vars) #remove_items
        remaining_vars = helpers.remove_items(all_variables, separate_vars)
        data_first = self._select_data(df=data, variables=separate_vars, column_name='boot_var')
        data_second = self._select_data(df=data, variables=remaining_vars, column_name='boot_var')

        fig, ax = plt.subplots(nrows=1, ncols=2,
                               gridspec_kw={'width_ratios': [len(separate_vars),
                                                             len(remaining_vars)
                                                             ]
                                            }
                               )
        if len(separate_vars) > 1:
            first_box_width = .8
        else:
            first_box_width = 2.

        sns.boxplot(x=self._x_name, y="data", hue="ahead", data=data_first, ax=ax[0], whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"},
                    flierprops={"marker": "."}, width=first_box_width
                    )
        ax[0].set(ylabel=f"skill score", xlabel="")

        sns.boxplot(x=self._x_name, y="data", hue="ahead", data=data_second, ax=ax[1], whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"},
                    flierprops={"marker": "."},
                    )
        ax[1].set(ylabel="", xlabel="")
        ax[1].yaxis.tick_right()
        handles, _ = ax[1].get_legend_handles_labels()
        for sax in ax:
            matplotlib.pyplot.sca(sax)
            sax.axhline(y=0, color="grey", linewidth=.5)
            plt.xticks(rotation=45, ha='right')
            sax.legend_.remove()

        fig.legend(handles, self._labels, loc='upper center', ncol=len(handles) + 1, )

        def align_yaxis(ax1, ax2):
            """
            Align zeros of the two axes, zooming them out by same ratio

            This function is copy pasted from https://stackoverflow.com/a/41259922
            """
            axes = (ax1, ax2)
            extrema = [ax.get_ylim() for ax in axes]
            tops = [extr[1] / (extr[1] - extr[0]) for extr in extrema]
            # Ensure that plots (intervals) are ordered bottom to top:
            if tops[0] > tops[1]:
                axes, extrema, tops = [list(reversed(l)) for l in (axes, extrema, tops)]

            # How much would the plot overflow if we kept current zoom levels?
            tot_span = tops[1] + 1 - tops[0]

            b_new_t = extrema[0][0] + tot_span * (extrema[0][1] - extrema[0][0])
            t_new_b = extrema[1][1] - tot_span * (extrema[1][1] - extrema[1][0])
            axes[0].set_ylim(extrema[0][0], b_new_t)
            axes[1].set_ylim(t_new_b, extrema[1][1])

        align_yaxis(ax[0], ax[1])
        align_yaxis(ax[0], ax[1])

    @staticmethod
    def _select_data(df: pd.DataFrame, variables: List[str], column_name: str) -> pd.DataFrame:
        for i, variable in enumerate(variables):
            if i == 0:
                selected_data = df.loc[df[column_name] == variable]
            else:
                tmp_var = df.loc[df[column_name] == variable]
                selected_data = pd.concat([selected_data, tmp_var], axis=0)
        return selected_data

    def raise_error_if_separate_vars_do_not_exist(self, data, separate_vars):
        if not self._variables_exist_in_df(df=data, variables=separate_vars):
            raise ValueError(f"At least one entry of `separate_vars' does not exist in `self.data' ")

    @staticmethod
    def _get_unique_values_from_column_of_df(df: pd.DataFrame, column_name: str) -> List:
        return list(df[column_name].unique())

    def _variables_exist_in_df(self, df: pd.DataFrame, variables: List[str], column_name: str = 'boot_var'):
        vars_in_df = set(self._get_unique_values_from_column_of_df(df, column_name))
        return set(variables).issubset(vars_in_df)

    def _plot_all_variables(self):
        """

        """
        fig, ax = plt.subplots()
        sns.boxplot(x=self._x_name, y="data", hue="ahead", data=self._data, ax=ax, whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, flierprops={"marker": "."})
        ax.axhline(y=0, color="grey", linewidth=.5)
        plt.xticks(rotation=45)
        ax.set(ylabel=f"skill score", xlabel="", title="summary of all stations")
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()


@TimeTrackingWrapper
class PlotTimeSeries:
    """
    Create time series plot.

    Currently, plots are under development and not well designed for any use in public.
    """

    def __init__(self, stations: List, data_path: str, name: str, window_lead_time: int = None, plot_folder: str = ".",
                 sampling="daily"):
        """Initialise."""
        self._data_path = data_path
        self._data_name = name
        self._stations = stations
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._sampling = self._get_sampling(sampling)
        self._plot(plot_folder)

    @staticmethod
    def _get_sampling(sampling):
        if sampling == "daily":
            return "D"
        elif sampling == "hourly":
            return "h"

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments.

        If window_lead_time is not given, extract this information from data itself by the number of ahead dimensions.
        If given, check if data supports the give length. If the number of ahead dimensions in data is lower than the
        given lead time, data's lead time is used.

        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._load_data(self._stations[0]).ahead)
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    def _load_data(self, station):
        logging.debug(f"... preprocess station {station}")
        file_name = os.path.join(self._data_path, self._data_name % station)
        data = xr.open_dataarray(file_name)
        return data.sel(type=["CNN", "obs"])

    def _plot(self, plot_folder):
        pdf_pages = self._create_pdf_pages(plot_folder)
        for pos, station in enumerate(self._stations):
            data = self._load_data(station)
            start, end = self._get_time_range(data)
            fig, axes, factor = self._create_subplots(start, end)
            nan_list = []
            for i_year in range(end - start + 1):
                data_year = data.sel(index=f"{start + i_year}")
                for i_half_of_year in range(factor):
                    pos = factor * i_year + i_half_of_year
                    plot_data = self._create_plot_data(data_year, factor, i_half_of_year)
                    self._plot_obs(axes[pos], plot_data)
                    self._plot_ahead(axes[pos], plot_data)
                    if np.isnan(plot_data.values).all():
                        nan_list.append(pos)
            self._clean_up_axes(nan_list, axes, fig)
            self._save_page(station, pdf_pages)
        pdf_pages.close()
        plt.close('all')

    @staticmethod
    def _clean_up_axes(nan_list, axes, fig):
        for i in reversed(nan_list):
            fig.delaxes(axes[i])

    @staticmethod
    def _save_page(station, pdf_pages):
        plt.suptitle(station)
        plt.legend()
        plt.tight_layout()
        pdf_pages.savefig(dpi=500)

    @staticmethod
    def _create_plot_data(data, factor, running_index):
        if factor > 1:
            if running_index == 0:
                data = data.where(data["index.month"] < 7)
            else:
                data = data.where(data["index.month"] >= 7)
        return data

    def _create_subplots(self, start, end):
        factor = 1
        if self._sampling == "h":
            factor = 2
        f, ax = plt.subplots((end - start + 1) * factor, sharey=True, figsize=(50, 30), squeeze=False)
        return f, ax[:, 0], factor

    def _plot_ahead(self, ax, data):
        color = sns.color_palette("Blues_d", self._window_lead_time).as_hex()
        for ahead in data.coords["ahead"].values:
            plot_data = data.sel(type="CNN", ahead=ahead).drop(["type", "ahead"]).squeeze().shift(index=ahead)
            label = f"{ahead}{self._sampling}"
            ax.plot(plot_data, color=color[ahead - 1], label=label)

    def _plot_obs(self, ax, data):
        ahead = 1
        obs_data = data.sel(type="obs", ahead=ahead).shift(index=ahead)
        # index = data.index + np.timedelta64(1, self._sampling)
        # ax.plot(index, obs_data.values, color=matplotlib.colors.cnames["green"], label="obs")
        ax.plot(obs_data, color=matplotlib.colors.cnames["green"], label="obs")

    @staticmethod
    def _get_time_range(data):
        def f(x, f_x):
            return pd.to_datetime(f_x(x.index.values)).year

        return f(data, min), f(data, max)

    @staticmethod
    def _create_pdf_pages(plot_folder: str):
        """
        Store plot locally.

        :param plot_folder: path to save the plot
        """
        plot_name = os.path.join(os.path.abspath(plot_folder), 'timeseries_plot.pdf')
        logging.debug(f"... save plot to {plot_name}")
        return matplotlib.backends.backend_pdf.PdfPages(plot_name)


@TimeTrackingWrapper
class PlotAvailability(AbstractPlotClass):
    """
    Create data availablility plot similar to Gantt plot.

    Each entry of given generator, will result in a new line in the plot. Data is summarised for given temporal
    resolution and checked whether data is available or not for each time step. This is afterwards highlighted as a
    colored bar or a blank space.

    You can set different colors to highlight subsets for example by providing different generators for the same index
    using different keys in the input dictionary.

    Note: each bar is surrounded by a small white box to highlight gabs in between. This can result in too long gabs
    in display, if a gab is only very short. Also this appears on a (fluent) transition from one to another subset.

    Calling this class will create three versions fo the availability plot.

    1) Data availability for each element
    1) Data availability as summary over all elements (is there at least a single elemnt for each time step)
    1) Combination of single and overall availability

    .. image:: ../../../../../_source/_plots/data_availability.png
        :width: 400

    .. image:: ../../../../../_source/_plots/data_availability_summary.png
        :width: 400

    .. image:: ../../../../../_source/_plots/data_availability_combined.png
        :width: 400

    """

    def __init__(self, generators: Dict[str, DataCollection], plot_folder: str = ".", sampling="daily",
                 summary_name="data availability", time_dimension="datetime"):
        """Initialise."""
        # create standard Gantt plot for all stations (currently in single pdf file with single page)
        super().__init__(plot_folder, "data_availability")
        self.dim = time_dimension
        self.sampling = self._get_sampling(sampling)
        self.linewidth = None
        if self.sampling == 'h':
            self.linewidth = 0.001
        plot_dict = self._prepare_data(generators)
        lgd = self._plot(plot_dict)
        self._save(bbox_extra_artists=(lgd,), bbox_inches="tight")
        # create summary Gantt plot (is data in at least one station available)
        self.plot_name += "_summary"
        plot_dict_summary = self._summarise_data(generators, summary_name)
        lgd = self._plot(plot_dict_summary)
        self._save(bbox_extra_artists=(lgd,), bbox_inches="tight")
        # combination of station and summary plot, last element is summary broken bar
        self.plot_name = "data_availability_combined"
        plot_dict_summary.update(plot_dict)
        lgd = self._plot(plot_dict_summary)
        self._save(bbox_extra_artists=(lgd,), bbox_inches="tight")

    def _prepare_data(self, generators: Dict[str, DataCollection]):
        plt_dict = {}
        for subset, data_collection in generators.items():
            for station in data_collection:
                labels = station.get_Y(as_numpy=False).resample({self.dim: self.sampling}, skipna=True).mean()
                labels_bool = labels.sel(window=1).notnull()
                group = (labels_bool != labels_bool.shift({self.dim: 1})).cumsum()
                plot_data = pd.DataFrame({"avail": labels_bool.values, "group": group.values},
                                         index=labels.coords[self.dim].values)
                t = plot_data.groupby("group").apply(lambda x: (x["avail"].head(1)[0], x.index[0], x.shape[0]))
                t2 = [i[1:] for i in t if i[0]]

                if plt_dict.get(str(station)) is None:
                    plt_dict[str(station)] = {subset: t2}
                else:
                    plt_dict[str(station)].update({subset: t2})
        return plt_dict

    def _summarise_data(self, generators: Dict[str, DataCollection], summary_name: str):
        plt_dict = {}
        for subset, data_collection in generators.items():
            all_data = None
            for station in data_collection:
                labels = station.get_Y(as_numpy=False).resample({self.dim: self.sampling}, skipna=True).mean()
                labels_bool = labels.sel(window=1).notnull()
                if all_data is None:
                    all_data = labels_bool
                else:
                    tmp = all_data.combine_first(labels_bool)  # expand dims to merged datetime coords
                    all_data = np.logical_or(tmp, labels_bool).combine_first(
                        all_data)  # apply logical on merge and fill missing with all_data

            group = (all_data != all_data.shift({self.dim: 1})).cumsum()
            plot_data = pd.DataFrame({"avail": all_data.values, "group": group.values},
                                     index=all_data.coords[self.dim].values)
            t = plot_data.groupby("group").apply(lambda x: (x["avail"].head(1)[0], x.index[0], x.shape[0]))
            t2 = [i[1:] for i in t if i[0]]
            if plt_dict.get(summary_name) is None:
                plt_dict[summary_name] = {subset: t2}
            else:
                plt_dict[summary_name].update({subset: t2})
        return plt_dict

    def _plot(self, plt_dict):
        colors = self.get_dataset_colors()
        pos = 0
        height = 0.8  # should be <= 1
        yticklabels = []
        number_of_stations = len(plt_dict.keys())
        fig, ax = plt.subplots(figsize=(10, number_of_stations / 3))
        for station, d in sorted(plt_dict.items(), reverse=True):
            pos += 1
            for subset, color in colors.items():
                plt_data = d.get(subset)
                if plt_data is None:
                    continue
                ax.broken_barh(plt_data, (pos, height), color=color, edgecolor="white", linewidth=self.linewidth)
            yticklabels.append(station)

        ax.set_ylim([height, number_of_stations + 1])
        ax.set_yticks(np.arange(len(plt_dict.keys())) + 1 + height / 2)
        ax.set_yticklabels(yticklabels)
        handles = [mpatches.Patch(color=c, label=k) for k, c in colors.items()]
        lgd = plt.legend(handles=handles, bbox_to_anchor=(0, 1, 1, 0.2), loc="lower center", ncol=len(handles))
        return lgd


@TimeTrackingWrapper
class PlotSeparationOfScales(AbstractPlotClass):

    def __init__(self, collection: DataCollection, plot_folder: str = "."):
        """Initialise."""
        # create standard Gantt plot for all stations (currently in single pdf file with single page)
        plot_folder = os.path.join(plot_folder, "separation_of_scales")
        super().__init__(plot_folder, "separation_of_scales")
        self._plot(collection)

    def _plot(self, collection: DataCollection):
        orig_plot_name = self.plot_name
        for dh in collection:
            data = dh.get_X(as_numpy=False)[0]
            station = dh.id_class.station[0]
            data = data.sel(Stations=station)
            # plt.subplots()
            data.plot(x="datetime", y="window", col="filter", row="variables", robust=True)
            self.plot_name = f"{orig_plot_name}_{station}"
            self._save()


@TimeTrackingWrapper
class PlotAvailabilityHistogram(AbstractPlotClass):
    """
    Create data availability plots as histogram.

    Each entry of each generator is checked for `notnull()` values along all the datetime axis (boolean).
    Calling this class creates two different types of histograms where each generator

    1) data_availability_histogram: datetime (xaxis) vs. number of stations with availabile data (yaxis)
    2) data_availability_histogram_cumulative: number of samples (xaxis) vs. number of stations having at least number
       of samples (yaxis)

    .. image:: ../../../../../_source/_plots/data_availability_histogram_hist.png
        :width: 400

    .. image:: ../../../../../_source/_plots/data_availability_histogram_hist_cum.png
        :width: 400

    """

    def __init__(self, generators: Dict[str, DataCollection], plot_folder: str = ".",
                 subset_dim: str = 'DataSet', history_dim: str = 'window',
                 station_dim: str = 'Stations',):

        super().__init__(plot_folder, "data_availability_histogram")

        self.subset_dim = subset_dim
        self.history_dim = history_dim
        self.station_dim = station_dim

        self.freq = None
        self.temporal_dim = None
        self.target_dim = None
        self._prepare_data(generators)

        for plt_type in self.allowed_plot_types:
            plot_name_tmp = self.plot_name
            self.plot_name += '_' + plt_type
            self._plot(plt_type=plt_type)
            self._save()
            self.plot_name = plot_name_tmp

    def _set_dims_from_datahandler(self, data_handler):
        self.temporal_dim = data_handler.id_class.time_dim
        self.target_dim = data_handler.id_class.target_dim
        self.freq = self._get_sampling(data_handler.id_class.sampling)

    @property
    def allowed_plot_types(self):
        plot_types = ['hist', 'hist_cum']
        return plot_types

    def _prepare_data(self, generators: Dict[str, DataCollection]):
        """
        Prepares data to be used by plot methods.

        Creates xarrays which are sums of valid data (boolean sums) across i) station_dim and ii) temporal_dim
        """
        avail_data_time_sum = {}
        avail_data_station_sum = {}
        dataset_time_interval = {}
        for subset, generator in generators.items():
            avail_list = []
            for station in generator:
                self._set_dims_from_datahandler(data_handler=station)
                station_data_x = station.get_X(as_numpy=False)[0]
                station_data_x = station_data_x.loc[{self.history_dim: 0,  # select recent window frame
                                                     self.target_dim: station_data_x[self.target_dim].values[0]}]
                station_data_x = self._reduce_dims(station_data_x)
                avail_list.append(station_data_x.notnull())
            avail_data = xr.concat(avail_list, dim=self.station_dim).notnull()
            avail_data_time_sum[subset] = avail_data.sum(dim=self.station_dim)
            avail_data_station_sum[subset] = avail_data.sum(dim=self.temporal_dim)
            dataset_time_interval[subset] = self._get_first_and_last_indexelement_from_xarray(
                avail_data_time_sum[subset], dim_name=self.temporal_dim, return_type='as_dict'
            )
        avail_data_amount = xr.concat(avail_data_time_sum.values(), pd.Index(avail_data_time_sum.keys(),
                                                                             name=self.subset_dim)
                                      )
        full_time_index = self._make_full_time_index(avail_data_amount.coords[self.temporal_dim].values, freq=self.freq)
        self.avail_data_cum_sum = xr.concat(avail_data_station_sum.values(), pd.Index(avail_data_station_sum.keys(),
                                                                                      name=self.subset_dim))
        self.avail_data_amount = avail_data_amount.reindex({self.temporal_dim: full_time_index})
        self.dataset_time_interval = dataset_time_interval

    def _reduce_dims(self, dataset):
        if len(dataset.dims) > 2:
            required = {self.temporal_dim, self.station_dim}
            unimportant = set(dataset.dims).difference(required)
            sel_dict = {un: dataset[un].values[0] for un in unimportant}
            dataset = dataset.loc[sel_dict]
        return dataset

    @staticmethod
    def _get_first_and_last_indexelement_from_xarray(xarray, dim_name, return_type='as_tuple'):
        if isinstance(xarray, xr.DataArray):
            first = xarray.coords[dim_name].values[0]
            last = xarray.coords[dim_name].values[-1]
            if return_type == 'as_tuple':
                return first, last
            elif return_type == 'as_dict':
                return {'first': first, 'last': last}
            else:
                raise TypeError(f"return_type must be 'as_tuple' or 'as_dict', but is '{return_type}'")
        else:
            raise TypeError(f"xarray must be of type xr.DataArray, but is of type {type(xarray)}")

    @staticmethod
    def _make_full_time_index(irregular_time_index, freq):
        full_time_index = pd.date_range(start=irregular_time_index[0], end=irregular_time_index[-1], freq=freq)
        return full_time_index

    def _plot(self, plt_type='hist', *args):
        if plt_type == 'hist':
            self._plot_hist()
        elif plt_type == 'hist_cum':
            self._plot_hist_cum()
        else:
            raise ValueError(f"plt_type mus be 'hist' or 'hist_cum', but is {type}")

    def _plot_hist(self, *args):
        colors = self.get_dataset_colors()
        fig, axes = plt.subplots(figsize=(10, 3))
        for i, subset in enumerate(self.dataset_time_interval.keys()):
            plot_dataset = self.avail_data_amount.sel({self.subset_dim: subset,
                                                       self.temporal_dim: slice(
                                                           self.dataset_time_interval[subset]['first'],
                                                           self.dataset_time_interval[subset]['last']
                                                       )
                                                       }
                                                      )

            plot_dataset.plot.step(color=colors[subset], ax=axes, label=subset)
            plt.fill_between(plot_dataset.coords[self.temporal_dim].values, plot_dataset.values, color=colors[subset])

        lgd = fig.legend(loc="upper right", ncol=len(self.dataset_time_interval),
                         facecolor='white', framealpha=1, edgecolor='black')
        for lgd_line in lgd.get_lines():
            lgd_line.set_linewidth(4.0)
        plt.gca().xaxis.set_major_locator(mdates.YearLocator())
        plt.title('')
        plt.ylabel('Number of samples')
        plt.tight_layout()

    def _plot_hist_cum(self, *args):
        colors = self.get_dataset_colors()
        fig, axes = plt.subplots(figsize=(10, 3))
        n_bins = int(self.avail_data_cum_sum.max().values)
        bins = np.arange(0, n_bins+1)
        descending_subsets = self.avail_data_cum_sum.max(dim=self.station_dim).sortby(
            self.avail_data_cum_sum.max(dim=self.station_dim), ascending=False
        ).coords[self.subset_dim].values

        for subset in descending_subsets:
            self.avail_data_cum_sum.sel({self.subset_dim: subset}).plot.hist(ax=axes,
                                                                             bins=bins,
                                                                             label=subset,
                                                                             cumulative=-1,
                                                                             color=colors[subset],
                                                                             # alpha=.5
                                                                             )

        lgd = fig.legend(loc="upper right", ncol=len(self.dataset_time_interval),
                         facecolor='white', framealpha=1, edgecolor='black')
        plt.title('')
        plt.ylabel('Number of stations')
        plt.xlabel('Number of samples')
        plt.xlim((bins[0], bins[-1]))
        plt.tight_layout()



if __name__ == "__main__":
    stations = ['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087']
    path = "../../testrun_network/forecasts"
    plt_path = "../../"

    con_quan_cls = PlotConditionalQuantiles(stations, path, plt_path)
